"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
exports.$listeners = Symbol('EventEmitter.listeners');
exports.$addListener = Symbol('EventEmitter.addListener');
/** Statically typed event emitter */
var EventEmitter = /** @class */ (function () {
    function EventEmitter() {
        this[exports.$listeners] = {};
    }
    /** Count the number of listeners for an event */
    EventEmitter.count = function (ee, key) {
        var count = 0;
        var list = ee[exports.$listeners][key];
        if (list) {
            var cb = list.first;
            while (++count) {
                if (cb.next) {
                    cb = cb.next;
                }
                else
                    break;
            }
        }
        return count;
    };
    /** Check if an event has listeners */
    EventEmitter.has = function (ee, key) {
        if (key == '*') {
            for (key in ee[exports.$listeners])
                return true;
            return false;
        }
        return ee[exports.$listeners][key] !== undefined;
    };
    /** Get an array of event keys that have listeners */
    EventEmitter.keys = function (ee) {
        return Object.keys(ee[exports.$listeners]);
    };
    /** Call the given listener when no other listeners exist */
    EventEmitter.unhandle = function (ee, key, impl) {
        var listener = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            if (!ee[exports.$listeners][key].first.next)
                return impl.apply(void 0, args);
        };
        return ee.on(key, listener);
    };
    /** Implementation */
    EventEmitter.prototype.on = function (arg, fn) {
        return this[exports.$addListener](arg, fn);
    };
    /** Implementation */
    EventEmitter.prototype.one = function (arg, fn) {
        return this[exports.$addListener](arg, fn, true);
    };
    /** Implementation */
    EventEmitter.prototype.off = function (arg, fn) {
        if (arg == '*') {
            var cache = this[exports.$listeners];
            this[exports.$listeners] = {};
            if (this._onEventUnhandled) {
                for (var key in cache) {
                    this._onEventUnhandled(key);
                }
            }
            return this;
        }
        if (typeof fn == 'function') {
            var list = this[exports.$listeners][arg];
            if (!list || unlink(list, function (l) { return l.fn == fn; })) {
                return this;
            }
        }
        delete this[exports.$listeners][arg];
        if (this._onEventUnhandled) {
            this._onEventUnhandled(arg);
        }
        return this;
    };
    /** Implementation */
    EventEmitter.prototype.emit = function (key) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        var result;
        var gen = this.listeners(key);
        while (true) {
            var _a = gen.next(), listener = _a.value, done = _a.done;
            if (done) {
                return result;
            }
            else {
                var generated = listener.apply(void 0, args);
                if (generated !== undefined) {
                    result = generated;
                }
            }
        }
    };
    /** Iterate over the listeners of an event */
    EventEmitter.prototype.listeners = function (key) {
        var list, prev, curr;
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    list = this[exports.$listeners][key];
                    if (!list)
                        return [2 /*return*/];
                    prev = null;
                    curr = list.first;
                    _a.label = 1;
                case 1:
                    if (!true) return [3 /*break*/, 3];
                    return [4 /*yield*/, curr.fn
                        // One-time listener
                    ];
                case 2:
                    _a.sent();
                    // One-time listener
                    if (curr.once) {
                        // Splice it.
                        if (prev) {
                            prev.next = curr.next;
                        }
                        // Shift it.
                        else if (curr.next) {
                            list.first = curr = curr.next;
                            return [3 /*break*/, 1];
                        }
                        // Delete it.
                        else {
                            delete this[exports.$listeners][key];
                            if (this._onEventUnhandled) {
                                this._onEventUnhandled(key);
                            }
                            return [2 /*return*/];
                        }
                    }
                    // Recurring listener
                    else {
                        prev = curr;
                    }
                    // Continue to the next listener.
                    if (curr.next) {
                        curr = curr.next;
                        return [3 /*break*/, 1];
                    }
                    // Update the last listener.
                    list.last = curr;
                    // All done.
                    return [2 /*return*/];
                case 3: return [2 /*return*/];
            }
        });
    };
    /** Implementation of the `on` and `one` methods */
    EventEmitter.prototype[exports.$addListener] = function (arg, fn, once) {
        if (once === void 0) { once = false; }
        var key = arg;
        if (fn) {
            var list = addListener(this[exports.$listeners], key, {
                fn: fn,
                once: once,
                next: null,
            });
            if (fn == list.first.fn && this._onEventHandled) {
                this._onEventHandled(key);
            }
            return fn;
        }
        if (typeof arg == 'object') {
            for (key in arg) {
                this[exports.$addListener](key, arg[key], once);
            }
        }
        return this;
    };
    /** Unique symbol for accessing the internal listener cache */
    EventEmitter.ev = exports.$listeners;
    return EventEmitter;
}());
exports.EventEmitter = EventEmitter;
function addListener(cache, key, cb) {
    var list = cache[key];
    if (list) {
        list.last.next = cb;
        list.last = cb;
    }
    else {
        cache[key] = list = { first: cb, last: cb };
    }
    return list;
}
/** Remove listeners that match the filter function */
function unlink(list, filter) {
    var prev = null;
    var curr = list.first;
    while (true) {
        // Return true to unlink the listener.
        if (filter(curr)) {
            // Splice it.
            if (prev) {
                prev.next = curr.next;
                if (curr.next) {
                    curr = curr.next;
                }
                else
                    break;
            }
            // Shift it.
            else if (curr.next) {
                list.first = curr = curr.next;
            }
            // No listeners remain.
            else {
                return null;
            }
        }
        // Keep this listener.
        else {
            prev = curr;
            if (curr.next) {
                curr = curr.next;
            }
            else
                break;
        }
    }
    // At least one listener remains.
    list.last = prev;
    return list;
}
//# sourceMappingURL=ee.js.map